package stamp.peripheral.web;
import stamp.core.*;
import stamp.errors.*;
import stamp.util.text.*;

/**
 * This class provides access to the Pink V2 (firmware 2) webserver module.
 * There are methods to read/write Pink variables, read status and post register,
 * check several conditions, commands to send UDP and email messages.
 *
 * History
 * V1.0 December 9, 2007 Initial release.
 * @author Peter Verkaik (verkaik6@zonnet.nl)
 * @version 1.0 December 9, 2007
 */
public class PinkV2 {

  //special read/write registers (in order of memory map)
  public static final int Nb_varBP = 100; //IP address of UDP message port
  public static final int Nb_varBM = 101; //content of udp message (both send and receive)
  public static final int Nb_varBI = 102; //IP destination address for udp message (send only)
  public static final int Nb_varEA = 103; //Email Authentication
  public static final int Nb_varEP = 104; //Email Password
  public static final int Nb_varEU = 105; //Email Username
  public static final int Nb_varEV = 106; //Email smtp server register
  public static final int Nb_varEC = 107; //Email Content register
  public static final int Nb_varES = 108; //Email Subject register
  public static final int Nb_varEF = 109; //Email From register
  public static final int Nb_varET = 110; //Email To register
  //readonly variables
  public static final int Nb_varSD = 111; //current DNS address register, read only
  public static final int Nb_varSG = 112; //current default gateway register, read only
  public static final int Nb_varSI = 113; //current IP address register, read only
  public static final int Nb_varSN = 114; //current network mask register, read only
  public static final int Nb_varST = 115; //STatus register, read only
  public static final int Nb_varSU = 116; //IP address of last incoming UDP message, read only
  public static final int Nb_varSV = 117; //post register, read only
  
  private static final int CLS = 0;

  private int id;
  private Uart tx;
  private Uart rx;
  private ScaledTimer16 t;

  /*
    Web Variable Range    Maximum Size   Address
    Nb_var00 to Nb_var19    64 bytes      0-19
    Nb_var20 to Nb_var99    64 bytes     20-99
    Nb_varBP                64 bytes       100
    Nb_varBM                64 bytes       101
    Nb_varBI                64 bytes       102
    Nb_varEA                64 bytes       103
    Nb_varEP                64 bytes       104
    Nb_varEU                64 bytes       105
    Nb_varEV                64 bytes       106
    Nb_varEC                64 bytes       107
    Nb_varES                64 bytes       108
    Nb_varEF                64 bytes       109
    Nb_varET                64 bytes       110
    Nb_varSD                16 bytes       111     readonly
    Nb_varSG                16 bytes       112     readonly
    Nb_varSI                16 bytes       113     readonly
    Nb_varSN                16 bytes       114     readonly
    Nb_varST                 1 byte        115     readonly
    Nb_varSU                16 bytes       116     readonly
    Nb_varSV                 1 byte        117     readonly
  */

  private static final char[] codes = new char[]{'B','P',  //100
                                                 'B','M',  //101
                                                 'B','I',  //102
                                                 'E','A',  //103
                                                 'E','P',  //104
                                                 'E','U',  //105
                                                 'E','V',  //106
                                                 'E','C',  //107
                                                 'E','S',  //108
                                                 'E','F',  //109
                                                 'E','T',  //110
                                                 'S','D',  //111 read only
                                                 'S','G',  //112 read only
                                                 'S','I',  //113 read only
                                                 'S','N',  //114 read only
                                                 'S','T',  //115 read only
                                                 'S','U',  //116 read only
                                                 'S','V'}; //117 read only

  private int status; //latest status
                      //b0 set if network link established
                      //b1 set if variable has been updated from webpage (cleared by status read)
                      //b2 set if Pink ready to send email
                      //b3 not used
                      //b4 set if email has been succesfully sent (cleared by status read or email failure)
                      //b5 set if udp message received (cleared by status read)
                      //b6 not used
                      //b7 not used

  /**
   * Constructor for Pink object
   *
   * @param id Identifies Pink webserver (use default '0')
   * @param tx Transmitter used to send commands and data to Pink
   * @param rx Receiver uaed to receive status and data from Pink
   * @param t Timer to timeout receiver
   */
  public PinkV2(int id, Uart tx, Uart rx, ScaledTimer16 t) {
    this.id = id;
    this.tx = tx;
    this.rx = rx;
    this.t = t;
  }

  /*
    The Pink Webserver follows standard AppMod structure and is formatted
    as follows
           !NB0<CMD><y>
    where
           !      Preface character alerts units on the bus that the message is an AppMod command
           NB     Pink Webserver Prefix (2 byte address denoting the type of AppMod addressed)
           0      Unit ID (Pink module to be addresses by this command, 0 is the default)
           <CMD> - Command, i.e. W = Write Variable, R = Read Variable, etc.
           <y>   - Command or configuration parameter

    To write:
              send !!','N','B',ID,'W',<xx>,':',<definition>,CLS
    where
             <xx> - Variable reference number from 00 to 99, or Email or UDP codes (see Sections 4 and 5).
                    For example, when defining Nb_var20, use 20.
               :  - Separator between the definition and the command.
     <definition> - Definition of variable
              CLS - This lets the Pink know where the definition ends. This is necessary
                    because it is possible to store less data than the variable is capable
                    of holding. The CLS character is used to let the Pink know that it has
                    reached the end of the data for that variable.
    To read:
              send '!','N','B',ID,'R',<xx>
    where
             <xx> - Variable reference number from 00 to 99. For example,
                    when reading Nb_var05, use 05.
  */

  /**
   * Clear range of Pink variables.
   *
   * @param addr Address (0-110) of first Pink variable to clear.
   * @param num Number of variables (1-111) to clear.
   */
  public void clearVar(int addr, int num) {
    if ((addr<0) || (addr>110) || (num<1)) return; //do not write to PINK with invalid parameters
    while ((addr<111) && (num>0)) {
      sendCommand(addr++,true);
      tx.sendByte(':');
      tx.sendByte(CLS); //inform PINK this write command is completed
      num--;
    }
  }

  /**
   * Write value to Pink variable.
   *
   * @param addr Address (0-106) of Pink variable.
   * @param str String holding value to write.
   * @return Number of bytes written (exclusive closing null) or errorcode.
   */
  public int writeVar(int addr, String str) {
    return writeVar(addr,str.toCharArray());
  }

  /**
   * Write value to Pink variable.
   * If the number of bytes to write is 64 or more, this method only writes the first 63 bytes.
   * If any of the bytes to write has the value CLS (binary 0) then the remaining bytes
   * are not written.
   * No bytes are written if any of the parameters is invalid.
   *
   * @param addr Address (0-110) of Pink variable.
   * @param buf Array holding value to write.
   * @return Number of bytes written (exclusive closing null) or errorcode.
   */
  public int writeVar(int addr, char[] buf) {
    int i,c;
    if ((addr<0) || (addr>110)) return err.INVALID; //do not write to PINK with invalid parameters
    int n = buf.length; //maximum number of bytes to write
    i = 0;
    sendCommand(addr,true);
    tx.sendByte(':');
    while (i < n) {
      c = buf[i];
      if (c != CLS) {
        tx.sendByte(c);
        i++;
      }
      else break;     //end of variable value
    }
    tx.sendByte(CLS); //inform PINK this write command is completed
    return i;
  }

  /**
   * Read value from Pink variable.
   * This method reads bytes from a variable until a CLS character is encountered,
   * or until the buffer is filled, whichever comes first. A closing 0 will be
   * appended to the bytes that are read.
   *
   * @param addr Address (0-117) of Pink variable.
   * @param buf Array to hold value (array size should be 64 bytes).
   * @return Number of bytes read (exclusive closing 0) or errorcode in case of timeout or invalid parameter.
   */
  public int readVar(int addr, char[] buf) {
    int i,c;
    if ((addr<0) || (addr>117)) return err.INVALID; //do not write to PINK with invalid parameters
    int n = buf.length - 1;          //maximum number of bytes to read, last byte is cleared
    buf[buf.length-1] = 0;
    sendCommand(addr,false);
    t.mark();
    i=0; c=-1;
    while ((c!=CLS) && !t.timeout()) { //keep reading until CLS, even if not stored in buf
      if (rx.byteAvailable()) {
        t.mark();
        c = rx.receiveByte();
        if (i<n) buf[i] = (char)c; //only write to buf until position is buf.length-1
        i++;
      }
    }
    return (c==CLS) ? i-1 : err.TIME_OUT;
  }

  /**
   * Send email message.
   * Nb_varET, Nb_varEF, Nb_varES, Nb_varEC and Nb_varEV must contain correct values.
   */
  public void sendEmail() {
    tx.sendByte('!');
    tx.sendByte('N');
    tx.sendByte('B');
    tx.sendByte(id);
    tx.sendByte('S');
    tx.sendByte('M');
  }

  /**
   * Send UDP message.
   * Nb_varBI and Nb_varBM must contain correct values.
   */
  public void sendUdp() {
    tx.sendByte('!');
    tx.sendByte('N');
    tx.sendByte('B');
    tx.sendByte(id);
    tx.sendByte('S');
    tx.sendByte('B');
  }

  /**
   * Get status register.
   * Also internally stored for status methods.
   *
   * @return status or errorcode
   */
  public int getStatus() {
    tx.sendByte('!');
    tx.sendByte('N');
    tx.sendByte('B');
    tx.sendByte(id);
    tx.sendByte('S');
    tx.sendByte('T');
    t.mark();
    while (!t.timeout()) {
      if (rx.byteAvailable()) {
        status = rx.receiveByte();
        return status;
      }
    }
    return err.TIME_OUT;
  }

  /**
   * Get post register.
   *
   * @return post or errorcode
   */
  public int getPost() {
    tx.sendByte('!');
    tx.sendByte('N');
    tx.sendByte('B');
    tx.sendByte(id);
    tx.sendByte('S');
    tx.sendByte('V');
    t.mark();
    while (!t.timeout()) {
      if (rx.byteAvailable()) {
        return rx.receiveByte();
      }
    }
    return err.TIME_OUT;
  }

  /**
   * Check if network connection is established.
   * Method getStatus() must have been called.
   *
   * @return True if network connection established, false otherwise.
   */
  public boolean isConnected() {
    return (status & 0x01) != 0;
  }

  /**
   * Check if variable is updated via webpage.
   * Method getStatus() must have been called.
   * Nb_varSV will hold the number of the last variable updated from a web page POST.
   * To read the value of this variable, use readVar(Nb_varSV,buf,offset).
   *
   * @return True if variable updated, false otherwise.
   */
  public boolean isUpdated() {
    return (status & 0x02) != 0;
  }

  /**
   * Check if Pink ready to send email.
   * Method getStatus() must have been called.
   * Nb_varET, Nb_varEF, Nb_varES, Nb_varEC and Nb_varEV must contain correct values.
   *
   * @return True if Pink ready to send email, false otherwise.
   */
  public boolean isReady() {
    return (status & 0x04) == 0;
  }

  /**
   * Check if email has been sent.
   * Method getStatus() must have been called.
   *
   * @return True if email has been sent, false otherwise.
   */
  public boolean isSent() {
    return (status & 0x10) != 0;
  }

  /**
   * Check if there is a new udp message.
   * Method getStatus() must have been called.
   *
   * @return True if new udp message received, false otherwise.
   */
  public boolean isUdp() {
    return (status & 0x20) != 0;
  }

  /**
   * Send command to Pink webserver.
   *
   * @param addr Address (0-117) of Pink variable.
   * @param write True for write command, false for read command
   */
  private void sendCommand(int addr, boolean write) {
    tx.sendByte('!');
    tx.sendByte('N');
    tx.sendByte('B');
    tx.sendByte(id);
    tx.sendByte(write ? 'W' : 'R');
    if (addr < 100) {
      tx.sendByte((addr/10)+'0');
      tx.sendByte((addr%10)+'0');
    }
    else {
      tx.sendByte(codes[(addr-100)<<1]);
      tx.sendByte(codes[((addr-100)<<1)+1]);
    }
  }

}